# sage.doctest: needs sage.combinat sage.modules
"""
Monomial symmetric functions
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2010 Anne Schilling <anne at math.ucdavis.edu> (addition)
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.libs.symmetrica.all as symmetrica
from sage.arith.misc import binomial, factorial, multinomial
from sage.combinat.partition import _Partitions
from sage.rings.infinity import infinity
from sage.rings.integer import Integer

from . import classical


class SymmetricFunctionAlgebra_monomial(classical.SymmetricFunctionAlgebra_classical):
    def __init__(self, Sym):
        """
        A class for methods related to monomial symmetric functions.

        INPUT:

        - ``self`` -- a monomial symmetric function basis
        - ``Sym`` -- an instance of the ring of the symmetric functions

        TESTS::

            sage: m = SymmetricFunctions(QQ).m()
            sage: m == loads(dumps(m))
            True
            sage: TestSuite(m).run(skip=['_test_associativity', '_test_distributivity', '_test_prod'])
            sage: TestSuite(m).run(elements = [m[1,1]+m[2], m[1]+2*m[1,1]])
        """
        classical.SymmetricFunctionAlgebra_classical.__init__(self, Sym, "monomial", 'm')

    def _dual_basis_default(self):
        """
        Return the default dual basis to ``self`` when no scalar product is specified.

        This method returns the dual basis of the monomial basis with
        respect to the standard scalar product, which is the
        homogeneous basis.

        EXAMPLES::

            sage: m = SymmetricFunctions(QQ).m()
            sage: h = SymmetricFunctions(QQ).h()
            sage: m.dual_basis() == h
            True

        TESTS::

            sage: m._dual_basis_default() is m.dual_basis()
            True
            sage: zee = lambda x : x.centralizer_size()
            sage: dm = m.dual_basis(zee)
            sage: dm[3,1].scalar(m[2,1,1])
            0
            sage: m[2,1,1].scalar(dm[3,1])
            0
        """
        return self.realization_of().h()

    def product(self, left, right):
        """
        Return the product of ``left`` and ``right``.

        - ``left``, ``right`` -- symmetric functions written in the
          monomial basis ``self``

        OUTPUT:

        - the product of ``left`` and ``right``, expanded in the
          monomial basis, as a dictionary whose keys are partitions and
          whose values are the coefficients of these partitions (more
          precisely, their respective monomial symmetric functions) in the
          product.

        EXAMPLES::

            sage: m = SymmetricFunctions(QQ).m()
            sage: a = m([2,1])
            sage: a^2
            4*m[2, 2, 1, 1] + 6*m[2, 2, 2] + 2*m[3, 2, 1] + 2*m[3, 3] + 2*m[4, 1, 1] + m[4, 2]

        ::

            sage: QQx.<x> = QQ['x']
            sage: m = SymmetricFunctions(QQx).m()
            sage: a = m([2,1])+x
            sage: 2*a # indirect doctest
            2*x*m[] + 2*m[2, 1]
            sage: a^2
            x^2*m[] + 2*x*m[2, 1] + 4*m[2, 2, 1, 1] + 6*m[2, 2, 2] + 2*m[3, 2, 1] + 2*m[3, 3] + 2*m[4, 1, 1] + m[4, 2]
        """
        # Use symmetrica to do the multiplication
        # A = left.parent()

        # Hack due to symmetrica crashing when both of the
        # partitions are the empty partition
        # if  R is ZZ or R is QQ:
        #     return symmetrica.mult_monomial_monomial(left, right)

        z_elt = {}
        for left_m, left_c in left._monomial_coefficients.items():
            for right_m, right_c in right._monomial_coefficients.items():

                # Hack due to symmetrica crashing when both of the
                # partitions are the empty partition
                if not left_m and not right_m:
                    z_elt[left_m] = left_c * right_c
                    continue

                d = symmetrica.mult_monomial_monomial({left_m: Integer(1)},
                                                      {right_m: Integer(1)}).monomial_coefficients()
                for m in d:
                    if m in z_elt:
                        z_elt[m] += left_c * right_c * d[m]
                    else:
                        z_elt[m] = left_c * right_c * d[m]
        return self._from_dict(z_elt)

    def from_polynomial(self, f, check=True):
        r"""
        Return the symmetric function in the monomial basis corresponding
        to the polynomial ``f``.

        INPUT:

        - ``self`` -- a monomial symmetric function basis
        - ``f`` -- a polynomial in finitely many variables over the
          same base ring as ``self``; it is assumed that this
          polynomial is symmetric
        - ``check`` -- boolean (default: ``True``); checks whether
          the polynomial is indeed symmetric

        OUTPUT:

        - This function converts a symmetric polynomial `f` in a
          polynomial ring in finitely many variables to a symmetric
          function in the monomial basis of the ring of symmetric
          functions over the same base ring.

        EXAMPLES::

            sage: m = SymmetricFunctions(QQ).m()
            sage: P = PolynomialRing(QQ, 'x', 3)
            sage: x = P.gens()
            sage: f = x[0] + x[1] + x[2]
            sage: m.from_polynomial(f)
            m[1]
            sage: f = x[0]**2+x[1]**2+x[2]**2
            sage: m.from_polynomial(f)
            m[2]
            sage: f = x[0]^2+x[1]
            sage: m.from_polynomial(f)
            Traceback (most recent call last):
            ...
            ValueError: x0^2 + x1 is not a symmetric polynomial
            sage: f = (m[2,1]+m[1,1]).expand(3)
            sage: m.from_polynomial(f)
            m[1, 1] + m[2, 1]
            sage: f = (2*m[2,1]+m[1,1]+3*m[3]).expand(3)
            sage: m.from_polynomial(f)
            m[1, 1] + 2*m[2, 1] + 3*m[3]

        """
        assert self.base_ring() == f.base_ring()
        if check and not f.is_symmetric():
            raise ValueError("%s is not a symmetric polynomial" % f)
        out = self._from_dict({_Partitions.element_class(_Partitions, list(e)): c
                               for e, c in f.monomial_coefficients().items()
                               if all(e[i+1] <= e[i] for i in range(len(e)-1))},
                              remove_zeros=False)
        return out

    def from_polynomial_exp(self, p):
        r"""
        Conversion from polynomial in exponential notation.

        INPUT:

        - ``self`` -- a monomial symmetric function basis
        - ``p`` -- a polynomial over the same base ring as ``self``

        OUTPUT:

        - This returns a symmetric function by mapping each monomial of
          `p` with exponents ``exp`` into `m_\lambda` where `\lambda` is
          the partition with exponential notation ``exp``.

        EXAMPLES::

            sage: m = SymmetricFunctions(QQ).m()
            sage: P = PolynomialRing(QQ,'x',5)
            sage: x = P.gens()

        The exponential notation of the partition `(5,5,5,3,1,1)` is::

            sage: Partition([5,5,5,3,1,1]).to_exp()
            [2, 0, 1, 0, 3]

        Therefore, the monomial::

            sage: f = x[0]^2 * x[2] * x[4]^3

        is mapped to::

            sage: m.from_polynomial_exp(f)
            m[5, 5, 5, 3, 1, 1]

        Furthermore, this function is linear::

            sage: f = 3 * x[3] + 2 * x[0]^2 * x[2] * x[4]^3
            sage: m.from_polynomial_exp(f)
            3*m[4] + 2*m[5, 5, 5, 3, 1, 1]

        .. SEEALSO::

            :func:`Partition`, :meth:`Partition.to_exp`
        """
        assert self.base_ring() == p.parent().base_ring()
        from sage.combinat.sf.sfa import _from_polynomial
        return _from_polynomial(p, self)

    def antipode_by_coercion(self, element):
        r"""
        The antipode of ``element`` via coercion to and from the power-sum
        basis or the Schur basis (depending on whether the power sums really
        form a basis over the given ground ring).

        INPUT:

        - ``element`` -- element in a basis of the ring of symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(QQ)
            sage: m = Sym.monomial()
            sage: m[3,2].antipode()
            m[3, 2] + 2*m[5]
            sage: m.antipode_by_coercion(m[3,2])
            m[3, 2] + 2*m[5]

            sage: Sym = SymmetricFunctions(ZZ)
            sage: m = Sym.monomial()
            sage: m[3,2].antipode()
            m[3, 2] + 2*m[5]
            sage: m.antipode_by_coercion(m[3,2])
            m[3, 2] + 2*m[5]

        .. TODO::

            Is there a not too difficult way to get the power-sum computations
            to work over any ring, not just one with coercion from `\QQ`?
        """
        from sage.rings.rational_field import RationalField
        if self.has_coerce_map_from(RationalField()):
            p = self.realization_of().powersum()
            return self(p.antipode(p(element)))

        s = self.realization_of().schur()
        return self(s.antipode(s(element)))

    def _magma_init_(self, magma):
        """
        Used in converting this ring to the corresponding ring in MAGMA.

        EXAMPLES::

            sage: # optional - magma
            sage: M = SymmetricFunctions(QQ).m()
            sage: t = 4*M[3,2]+9
            sage: mt = magma(t); mt
            9 + 4*$.[3,2]
            sage: mt.sage()
            9*m[] + 4*m[3, 2]
        """
        B = magma(self.base_ring())
        Bref = B._ref()
        return f"SymmetricFunctionAlgebraMonomial({Bref})"

    class Element(classical.SymmetricFunctionAlgebra_classical.Element):
        def expand(self, n, alphabet='x'):
            """
            Expand the symmetric function ``self`` as a symmetric polynomial
            in ``n`` variables.

            INPUT:

            - ``n`` -- nonnegative integer

            - ``alphabet`` -- (default: ``'x'``) a variable for the expansion

            OUTPUT:

            A monomial expansion of ``self`` in the `n` variables
            labelled by ``alphabet``.

            EXAMPLES::

                sage: m = SymmetricFunctions(QQ).m()
                sage: m([2,1]).expand(3)
                x0^2*x1 + x0*x1^2 + x0^2*x2 + x1^2*x2 + x0*x2^2 + x1*x2^2
                sage: m([1,1,1]).expand(2)
                0
                sage: m([2,1]).expand(3,alphabet='z')
                z0^2*z1 + z0*z1^2 + z0^2*z2 + z1^2*z2 + z0*z2^2 + z1*z2^2
                sage: m([2,1]).expand(3,alphabet='x,y,z')
                x^2*y + x*y^2 + x^2*z + y^2*z + x*z^2 + y*z^2
                sage: m([1]).expand(0)
                0
                sage: (3*m([])).expand(0)
                3
            """

            def condition(part):
                return len(part) > n
            return self._expand(condition, n, alphabet)

        def principal_specialization(self, n=infinity, q=None):
            r"""
            Return the principal specialization of a symmetric function.

            The *principal specialization* of order `n` at `q`
            is the ring homomorphism `ps_{n,q}` from the ring of
            symmetric functions to another commutative ring `R`
            given by `x_i \mapsto q^{i-1}` for `i \in \{1,\dots,n\}`
            and `x_i \mapsto 0` for `i > n`.
            Here, `q` is a given element of `R`, and we assume that
            the variables of our symmetric functions are
            `x_1, x_2, x_3, \ldots`.
            (To be more precise, `ps_{n,q}` is a `K`-algebra
            homomorphism, where `K` is the base ring.)
            See Section 7.8 of [EnumComb2]_.

            The *stable principal specialization* at `q` is the ring
            homomorphism `ps_q` from the ring of symmetric functions
            to another commutative ring `R` given by
            `x_i \mapsto q^{i-1}` for all `i`.
            This is well-defined only if the resulting infinite sums
            converge; thus, in particular, setting `q = 1` in the
            stable principal specialization is an invalid operation.

            INPUT:

            - ``n`` -- (default: ``infinity``) a nonnegative integer or
              ``infinity``, specifying whether to compute the principal
              specialization of order ``n`` or the stable principal
              specialization.

            - ``q`` -- (default: ``None``) the value to use for `q`; the
              default is to create a ring of polynomials in ``q``
              (or a field of rational functions in ``q``) over the
              given coefficient ring.

            For ``q=1`` and finite ``n`` we use the formula from
            Proposition 7.8.3 of [EnumComb2]_:

            .. MATH::

                ps_{n,1}(m_\lambda) = \binom{n}{\ell(\lambda)}
                                      \binom{\ell(\lambda)}{m_1(\lambda), m_2(\lambda),\dots},

            where `\ell(\lambda)` denotes the length of `\lambda`.

            In all other cases, we convert to complete homogeneous
            symmetric functions.

            EXAMPLES::

                sage: m = SymmetricFunctions(QQ).m()
                sage: x = m[3,1]
                sage: x.principal_specialization(3)
                q^7 + q^6 + q^5 + q^3 + q^2 + q

                sage: x = 5*m[2] + 3*m[1] + 1
                sage: x.principal_specialization(3, q=var("q"))                         # needs sage.symbolic
                -10*(q^3 - 1)*q/(q - 1) + 5*(q^3 - 1)^2/(q - 1)^2 + 3*(q^3 - 1)/(q - 1) + 1

            TESTS::

                sage: m.zero().principal_specialization(3)
                0
            """
            if n == 1:
                R = self.base_ring()
                mc = self.monomial_coefficients(copy=False).items()
                return R.sum(c for partition, c in mc
                             if len(partition) <= 1)

            if q == 1:
                if n == infinity:
                    raise ValueError("the stable principal specialization at q=1 is not defined")
                f = lambda partition: binomial(n, len(partition))*multinomial(partition.to_exp())
                return self.parent()._apply_module_morphism(self, f, q.parent())

            # heuristically, it seems fastest to fall back to the
            # elementary basis - using the powersum basis would
            # introduce singularities, because it is not a Z-basis
            return self.parent().realization_of().elementary()(self).principal_specialization(n=n, q=q)

        def exponential_specialization(self, t=None, q=1):
            r"""
            Return the exponential specialization of a
            symmetric function (when `q = 1`), or the
            `q`-exponential specialization (when `q \neq 1`).

            The *exponential specialization* `ex` at `t` is a
            `K`-algebra homomorphism from the `K`-algebra of
            symmetric functions to another `K`-algebra `R`.
            It is defined whenever the base ring `K` is a
            `\QQ`-algebra and `t` is an element of `R`.
            The easiest way to define it is by specifying its
            values on the powersum symmetric functions to be
            `p_1 = t` and `p_n = 0` for `n > 1`.
            Equivalently, on the homogeneous functions it is
            given by `ex(h_n) = t^n / n!`; see Proposition 7.8.4 of
            [EnumComb2]_.

            By analogy, the `q`-exponential specialization is a
            `K`-algebra homomorphism from the `K`-algebra of
            symmetric functions to another `K`-algebra `R` that
            depends on two elements `t` and `q` of `R` for which
            the elements `1 - q^i` for all positive integers `i`
            are invertible.
            It can be defined by specifying its values on the
            complete homogeneous symmetric functions to be

            .. MATH::

                ex_q(h_n) = t^n / [n]_q!,

            where `[n]_q!` is the `q`-factorial.  Equivalently, for
            `q \neq 1` and a homogeneous symmetric function `f` of
            degree `n`, we have

            .. MATH::

                ex_q(f) = (1-q)^n t^n ps_q(f),

            where `ps_q(f)` is the stable principal specialization of `f`
            (see :meth:`principal_specialization`).
            (See (7.29) in [EnumComb2]_.)

            The limit of `ex_q` as `q \to 1` is `ex`.

            INPUT:

            - ``t`` -- (default: ``None``) the value to use for `t`;
              the default is to create a ring of polynomials in ``t``

            - ``q`` -- (default: `1`) the value to use for `q`;  if
              ``q`` is ``None``, then a ring (or fraction field) of
              polynomials in ``q`` is created

            EXAMPLES::

                sage: m = SymmetricFunctions(QQ).m()
                sage: (m[3]+m[2,1]+m[1,1,1]).exponential_specialization()
                1/6*t^3

                sage: x = 5*m[1,1,1] + 3*m[2,1] + 1
                sage: x.exponential_specialization()
                5/6*t^3 + 1

            We also support the `q`-exponential_specialization::

                sage: factor(m[3].exponential_specialization(q=var("q"), t=var("t")))   # needs sage.symbolic
                (q - 1)^2*t^3/(q^2 + q + 1)

            TESTS::

                sage: m.zero().exponential_specialization()
                0
            """
            def get_variable(ring, name):
                try:
                    ring(name)
                except TypeError:
                    from sage.rings.polynomial.polynomial_ring_constructor import (
                        PolynomialRing,
                    )
                    return PolynomialRing(ring, name).gen()
                else:
                    raise ValueError("the variable %s is in the base ring, pass it explicitly" % name)

            if q == 1:
                if t is None:
                    t = get_variable(self.base_ring(), 't')

                def f(partition):
                    n = 0
                    for part in partition:
                        if part != 1:
                            return 0
                        n += 1
                    return t**n / factorial(n)

                return self.parent()._apply_module_morphism(self, f, t.parent())

            # heuristically, it seems fastest to fall back to the
            # elementary basis - using the powersum basis would
            # introduce singularities, because it is not a Z-basis
            return self.parent().realization_of().elementary()(self).exponential_specialization(t=t, q=q)


# Backward compatibility for unpickling
from sage.misc.persist import register_unpickle_override

register_unpickle_override('sage.combinat.sf.monomial', 'SymmetricFunctionAlgebraElement_monomial', SymmetricFunctionAlgebra_monomial.Element)



r"""
Affine and Projective Plane Curve Arrangements

We create classes :class:`AffinePlaneCurveArrangements`
and :class:`ProjectivePlaneCurveArrangements`
following the properties of :class:`HyperplaneArrangements`::

    sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
    sage: C = H(3*x + 2*y - x^2 + y^3 - 7);  C
    Arrangement (y^3 - x^2 + 3*x + 2*y - 7) in Affine Space of dimension 2 over Rational Field

The individual curves will be in  :class:`AffinePlaneCurve` or in :class:`ProjectivePlaneCurve`::

    sage: C[0].parent()
    <class 'sage.schemes.curves.affine_curve.IntegralAffinePlaneCurve_with_category'>

The default base field is `\QQ`, the rational numbers.
Number fields are also possible (also with fixed embeddings in
`\QQbar`)::

    sage: # needs sage.rings.number_field
    sage: x = polygen(QQ, 'x')
    sage: NF.<a> = NumberField(x^4 - 5 * x^2 + 5, embedding=1.90)
    sage: H.<y,z> = AffinePlaneCurveArrangements(NF)
    sage: A = H(y^2 - a * z, y^2 + a * z); A
    Arrangement (y^2 + (-a)*z, y^2 + a*z) in Affine Space of dimension 2
    over Number Field in a with defining polynomial
    x^4 - 5*x^2 + 5 with a = 1.902113032590308?
    sage: A.base_ring()
    Number Field in a with defining polynomial x^4 - 5*x^2 + 5
    with a = 1.902113032590308?


AUTHORS:

- Enrique Artal (2023-10): initial version
"""

# *****************************************************************************
#       Copyright (C) 2023 Enrique Artal <artal@unizar.es>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from itertools import combinations
from typing import Any

from sage.categories.sets_cat import Sets
from sage.groups.free_group import FreeGroup
from sage.misc.abstract_method import abstract_method
from sage.misc.cachefunc import cached_method
from sage.misc.misc_c import prod
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.qqbar import QQbar
from sage.rings.ring import _Fields
from sage.schemes.affine.affine_space import AffineSpace
from sage.schemes.curves.affine_curve import AffinePlaneCurve
from sage.schemes.curves.constructor import Curve
from sage.schemes.curves.projective_curve import ProjectiveSpace, ProjectivePlaneCurve
from sage.schemes.curves.zariski_vankampen import braid_monodromy, fundamental_group_arrangement
from sage.structure.category_object import normalize_names
from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.structure.richcmp import richcmp
from sage.structure.unique_representation import UniqueRepresentation


class PlaneCurveArrangementElement(Element):
    """
    An ordered plane curve arrangement.
    """
    def __init__(self, parent, curves, check=True) -> None:
        """
        Construct a plane curve arrangement.

        INPUT:

        - ``parent`` -- the parent :class:`PlaneCurveArrangements`

        - ``curves`` -- tuple of curves

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: elt = H(x, y); elt
            Arrangement (x, y) in Affine Space of dimension 2 over Rational Field
            sage: TestSuite(elt).run()
            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: elt = H(x, y); elt
            Arrangement (x, y) in Projective Space of dimension 2 over Rational Field
            sage: TestSuite(elt).run()
        """
        super().__init__(parent)
        self._curves = tuple(curves)
        if check:
            affine = all(isinstance(h, AffinePlaneCurve) for h in curves)
            projective = all(isinstance(h, ProjectivePlaneCurve) for h in curves)
            if not (affine or projective):
                raise ValueError("not all elements are curves")
            if not all(h.ambient_space() is parent.ambient_space()
                       for h in curves):
                raise ValueError("not all curves are in the same ambient space")

    def __getitem__(self, i):
        """
        Return the `i`-th curve.

        INPUT:

        - ``i`` -- integer

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: H(y^2 - x, y^3 + 2 * x^2, x^4 + y^4 + 1)
            Arrangement (y^2 - x, y^3 + 2*x^2, x^4 + y^4 + 1)
            in Affine Space of dimension 2 over Rational Field
        """
        return self._curves[i]

    def __hash__(self) -> int:
        r"""
        TESTS::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: H((x * y, x + y +1)).__hash__()   # random
            -4938643871296220686
        """
        return hash(self.curves())

    def ncurves(self) -> int:
        r"""
        Return the number of curves in the arrangement.

        OUTPUT: integer

        EXAMPLES::

            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: h = H((x * y, x + y + z))
            sage: h.ncurves()
            2
            sage: len(h)    # equivalent
            2
        """
        return len(self._curves)

    __len__ = ncurves

    def curves(self) -> tuple:
        r"""
        Return the curves in the arrangement as a tuple.

        OUTPUT: a tuple

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: h = H((x * y, x + y + 1))
            sage: h.curves()
            (Affine Plane Curve over Rational Field defined by x*y,
             Affine Plane Curve over Rational Field defined by x + y + 1)

        Note that the curves can be indexed as if they were a list::

            sage: h[1]
            Affine Plane Curve over Rational Field defined by x + y + 1
        """
        return self._curves

    def _repr_(self) -> str:
        r"""
        String representation for a curve arrangement.

        OUTPUT: string

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: h = H([x * y, x + y + 1, x^3 - y^5, x^2 * y^2 + x^5 + y^5, (x^2 + y^2)^3 + (x^3 + y^3 - 1)^2])
            sage: h
            Arrangement of 5 curves in Affine Space of dimension 2 over Rational Field
            sage: H(())
            Arrangement () in Affine Space of dimension 2 over Rational Field
            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: h = H([x * y, x + y + z, x^3 * z^2 - y^5, x^2 * y^2 * z + x^5 + y^5, (x^2 + y^2)^3 + (x^3 + y^3 - z^3)^2])
            sage: h
            Arrangement of 5 curves in Projective Space of dimension 2 over Rational Field
        """
        if not self:
            return 'Empty curve arrangement in {}'.format(self.parent().ambient_space())
        if len(self) < 5:
            curves = ', '.join(h.defining_polynomial()._repr_()
                               for h in self._curves)
            return 'Arrangement ({}) in {}'.format(curves,
                                                   self.parent().ambient_space())
        return 'Arrangement of {} curves in {}'.format(len(self),
                                                       self.parent().ambient_space())

    def _richcmp_(self, other, op) -> bool:
        """
        Compare two curve arrangements.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: H(x) == H(y)
            False
            sage: H(x) == H(2 * x)
            True

        TESTS::

            sage: H(x) == 0
            False
        """
        return richcmp(self._curves, other._curves, op)

    def union(self, other):
        r"""
        The union of ``self`` with ``other``.

        INPUT:

        - ``other`` -- a curve arrangement or something that can
          be converted into a curve arrangement

        OUTPUT: a new curve arrangement

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: h = H([x * y, x + y + 1, x^3 - y^5, x^2 * y^2 + x^5 + y^5, (x^2 + y^2)^3 + (x^3 + y^3 - 1)^2])
            sage: C = Curve(x^8 - y^8 -x^4 * y^4)
            sage: h1 = h.union(C); h1
            Arrangement of 6 curves in Affine Space of dimension 2 over Rational Field
            sage: h1 == h1.union(C)
            True
        """
        P = self.parent()
        other_h = P(other)
        curves0 = self._curves + other_h._curves
        curves = []
        for h in curves0:
            if h not in curves:
                curves.append(h)
        result = P(*curves)
        return result

    add_curves = union

    __or__ = union

    def deletion(self, curves):
        r"""
        Return the curve arrangement obtained by removing ``curves``.

        INPUT:

        - ``curves`` -- a curve or curve arrangement

        OUTPUT:

        A new curve arrangement with the given curve(s)
        ``h`` removed.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: h = H([x * y, x + y + 1, x^3 - y^5, x^2 * y^2 + x^5 + y^5, (x^2 + y^2)^3 + (x^3 + y^3 - 1)^2])
            sage: C = h[-1]
            sage: h.deletion(C)
            Arrangement (x*y, x + y + 1, -y^5 + x^3, x^5 + y^5 + x^2*y^2)
            in Affine Space of dimension 2 over Rational Field
            sage: h.deletion(x)
            Traceback (most recent call last):
            ...
            ValueError: curve is not in the arrangement
            """
        parent = self.parent()
        curves = parent(curves)
        planes = list(self)
        for curve in curves:
            try:
                planes.remove(curve)
            except ValueError:
                raise ValueError('curve is not in the arrangement')
        return parent(planes)

    def change_ring(self, base_ring):
        """
        Return curve arrangement over the new base ring.

        INPUT:

        - ``base_ring`` -- the new base ring; must be a field for
          curve arrangements

        OUTPUT:

        The curve arrangement obtained by changing the base
        field, as a new curve arrangement.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 - x^3, x, y, y^2 + x * y + x^2)
            sage: K.<a> = CyclotomicField(3)
            sage: A.change_ring(K)
            Arrangement (-x^3 + y^2, x, y, x^2 + x*y + y^2) in Affine Space of
            dimension 2 over Cyclotomic Field of order 3 and degree 2
        """
        parent = self.parent().change_ring(base_ring)
        curves = tuple(c.change_ring(base_ring) for c in self)
        return parent(curves)

    @cached_method
    def coordinate_ring(self):
        """
        Return the coordinate ring of ``self``.

        OUTPUT: the coordinate ring of the curve arrangement

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: C = L(x, y)
            sage: C.coordinate_ring()
            Multivariate Polynomial Ring in x, y over Rational Field
            sage: P.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: C = P(x, y)
            sage: C.coordinate_ring()
            Multivariate Polynomial Ring in x, y, z over Rational Field
        """
        return self._curves[0].defining_polynomial().parent()

    def defining_polynomials(self) -> tuple:
        r"""
        Return the defining polynomials of the elements of ``self``.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 - x^3, x, y, y^2 + x * y + x^2)
            sage: A.defining_polynomials()
            (-x^3 + y^2, x, y, x^2 + x*y + y^2)
        """
        return tuple([h.defining_polynomial() for h in self._curves])

    def defining_polynomial(self, simplified=True):
        r"""
        Return the defining polynomial of the union of the curves in ``self``.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y ** 2 + x ** 2, x, y)
            sage: prod(A.defining_polynomials()) == A.defining_polynomial()
            True
        """
        return prod(self.defining_polynomials())

    def have_common_factors(self) -> bool:
        r"""
        Check if the curves have common factors.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(x * y, x^2 + x* y^3)
            sage: A.have_common_factors()
            True
            sage: H(x * y, x + y^3).have_common_factors()
            False
        """
        L = [c.defining_polynomial() for c in self._curves]
        C = combinations(L, 2)
        return any(f1.gcd(f2).degree() > 0 for f1, f2 in C)

    def reduce(self, clean=False, verbose=False):
        r"""
        Replace the curves by their reduction.

        INPUT:

        - ``clean`` -- boolean (default: ``False``); if ``False``
          and there are common factors it returns ``None`` and
          a warning message. If ``True``, the common factors are kept
          only in the first occurrence.

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2, (x + y)^3 * (x^2 + x * y + y^2))
            sage: A.reduce()
            Arrangement (y, x^3 + 2*x^2*y + 2*x*y^2 + y^3) in Affine Space
            of dimension 2 over Rational Field
            sage: C = H(x*y, x*(y + 1))
            sage: C.reduce(verbose=True)
            Some curves have common components
            sage: C.reduce(clean=True)
            Arrangement (x*y, y + 1) in Affine Space of dimension 2
            over Rational Field
            sage: C = H(x*y, x)
            sage: C.reduce(clean=True)
            Arrangement (x*y) in Affine Space of dimension 2 over Rational Field
        """
        P = self.parent()
        L = [self._curves[0].defining_polynomial().radical()]
        for c in self._curves[1:]:
            g = c.defining_polynomial().radical()
            for f in L:
                d = g.gcd(f)
                if d.degree() > 0 and not clean:
                    if verbose:
                        print("Some curves have common components")
                    return None
                g //= d
            if g.degree() > 0:
                L.append(g)
        return P(*L)


class AffinePlaneCurveArrangementElement(PlaneCurveArrangementElement):
    """
    An ordered affine plane curve arrangement.
    """
    def __init__(self, parent, curves, check=True) -> None:
        """
        Construct an ordered affine plane curve arrangement.

        INPUT:

        - ``parent`` -- the parent :class:`AffinePlaneCurveArrangements`

        - ``curves`` -- tuple of curves

        EXAMPLES::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: elt = H(x, y); elt
            Arrangement (x, y) in Affine Space of dimension 2 over Rational Field
            sage: TestSuite(elt).run()
        """
        Element.__init__(self, parent)
        self._curves = tuple(curves)
        if check:
            if not all(isinstance(h, AffinePlaneCurve) for h in curves):
                raise ValueError("not all elements are curves")
            if not all(h.ambient_space() is self.parent().ambient_space()
                       for h in curves):
                raise ValueError("not all curves are in the same ambient space")
        self._braid_monodromy_non_vertical = None
        self._braid_monodromy_vertical = None
        self._strands_nonvertical = None
        self._strands_vertical = None
        self._fundamental_group_nonsimpl_nonvertical = None
        self._fundamental_group_nonsimpl_vertical = None
        self._fundamental_group_simpl_nonvertical = None
        self._fundamental_group_simpl_vertical = None
        self._meridians_nonsimpl_nonvertical = None
        self._meridians_nonsimpl_vertical = None
        self._meridians_simpl_nonvertical = None
        self._meridians_simpl_vertical = None
        self._vertical_lines_in_braid_mon = None

    def fundamental_group(self, simplified=True, vertical=True,
                          projective=False):
        r"""
        Return the fundamental group of the complement of the union
        of affine plane curves in `\CC^2`.

        INPUT:

        - ``vertical`` -- boolean (default: ``True``); if ``True``, there
          are no vertical asymptotes, and there are vertical lines, then a
          simplified braid :func:`braid_monodromy` is used

        - ``simplified`` -- boolean (default: ``True``); if it is ``True``, the
          group is simplified

        - ``projective`` -- boolean (default: ``False``); to be used in the
          method for projective curves

        OUTPUT: a finitely presented group

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x, y + x - 1, x)
            sage: A.fundamental_group()
            Finitely presented group
            < x0, x1, x2 | x2*x0*x2^-1*x0^-1, x1*x0*x1^-1*x0^-1, (x2*x1)^2*(x2^-1*x1^-1)^2 >
            sage: A.meridians()
            {0: [x1, x2*x1*x2^-1], 1: [x0], 2: [x2],
             3: [x1^-1*x2^-1*x1^-1*x0^-1]}
            sage: G = A.fundamental_group(simplified=False)
            sage: G.sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3 | x3^-1*x2^-1*x3*x0*x1*x0^-1,
                               x3^-1*x1^-1*x3*x0*x1*x0^-1*x2^-1*x0^-1*(x2*x0)^2*x1^-1*x0^-1,
                               x3^-1*x0^-1*x3*x0*x1*x0^-1*x2^-1*x0*x2*x0*x1^-1*x0^-1,
                               x2^-1*x0^-1*x2*x0, x1^-1*x0^-1*x1*x0 >
            sage: A.meridians(simplified=False)
            {0: [x1, x2], 1: [x0], 2: [x3], 3: [x3^-1*x2^-1*x1^-1*x0^-1]}
            sage: A.fundamental_group(vertical=False)
            Finitely presented group
            < x0, x1, x2 | x2*x1^-1*x2^-1*x1, x1*x0*x1^-1*x0^-1, (x0*x2)^2*(x0^-1*x2^-1)^2 >
            sage: A.meridians(vertical=False)
            {0: [x2, x0*x2*x0^-1], 1: [x1], 2: [x0], 3: [x0*x2^-1*x0^-1*x2^-1*x1^-1*x0^-1]}
            sage: G = A.fundamental_group(simplified=False, vertical=False)
            sage: G.sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3 | x3^-1*x2^-1*x1^-1*x2*x3*x2^-1*x1*x2,
                               x3^-1*x2^-1*x1^-1*x2*x3*x2^-1*x1*x2,
                               (x3^-1*x2^-1*x0^-1*x2)^2*(x3*x2^-1*x0*x2)^2,
                               x3^-1*x2^-1*x0^-1*x2*x3^-1*x2^-1*x0*x2*x3*x2,
                               x1^-1*x0^-1*x1*x0 >
            sage: A.meridians(simplified=False, vertical=False)
            {0: [x2, x3], 1: [x1], 2: [x0], 3: [x3^-1*x2^-1*x1^-1*x0^-1]}
            sage: A = H(x * y^2 + x + y, y + x -1, x, y)
            sage: G = A.fundamental_group()  # long time (:issue:`39569`)
            sage: G.sorted_presentation()  # long time (:issue:`39569`)
            Finitely presented group
            < x0, x1, x2, x3 | x3^-1*x2^-1*x3*x2, x3^-1*x1^-1*x3*x1,
                               x3^-1*x0^-1*x3*x0, x2^-1*x1^-1*x2*x1,
                               x2^-1*x0^-1*x2*x0, x1^-1*x0^-1*x1*x0 >
        """
        if simplified and vertical:
            computed = self._fundamental_group_simpl_vertical
        elif simplified and not vertical:
            computed = self._fundamental_group_simpl_nonvertical
        elif not simplified and vertical:
            computed = self._fundamental_group_nonsimpl_vertical
        else:
            computed = self._fundamental_group_nonsimpl_nonvertical
        if computed:
            return computed
        K = self.base_ring()
        R = self.coordinate_ring()
        if not K.is_subring(QQbar):
            raise TypeError('the base field is not in QQbar')
        C = self.reduce()
        L = C.defining_polynomials()
        if vertical:
            bm = self._braid_monodromy_vertical
        else:
            bm = self._braid_monodromy_non_vertical
        if bm is not None:  # bm could be []
            if not vertical:
                st = self._strands_nonvertical
                d1 = prod(L).degree()
                bd = (bm, st, {}, d1)
            else:
                st = self._strands_vertical
                d1 = prod(L).degree(R.gen(1))
                bd = (bm, st, self._vertical_lines_in_braid_mon, d1)
        else:
            bd = None
        G, dic = fundamental_group_arrangement(L, simplified=simplified,
                                               puiseux=True,
                                               projective=projective,
                                               vertical=vertical,
                                               braid_data=bd)
        if simplified and vertical:
            self._fundamental_group_simpl_vertical = G
            self._meridians_simpl_vertical = dic
        elif simplified and not vertical:
            self._fundamental_group_simpl_nonvertical = G
            self._meridians_simpl_nonvertical = dic
        elif not simplified and vertical:
            self._fundamental_group_nonsimpl_vertical = G
            self._meridians_nonsimpl_vertical = dic
        else:
            self._fundamental_group_nonsimpl_nonvertical = G
            self._meridians_nonsimpl_nonvertical = dic
        return G

    def meridians(self, simplified=True, vertical=True) -> dict:
        r"""
        Return the meridians of each irreducible component.

        OUTPUT:

        A dictionary which associates the index of each curve with its meridians,
        including the line at infinity if it can be omputed

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed
           and :meth:`AffinePlaneCurveArrangements.fundamental_group` with the same options,
           where some examples are shown.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(x-1, y, x, y - 1)
            sage: A.fundamental_group()
            Finitely presented group
            < x0, x1, x2, x3 | x2*x0*x2^-1*x0^-1, x2*x1*x2^-1*x1^-1,
                               x3*x0*x3^-1*x0^-1, x3*x1*x3^-1*x1^-1 >
            sage: A.meridians()
            {0: [x2], 1: [x0], 2: [x3], 3: [x1], 4: [x3^-1*x2^-1*x1^-1*x0^-1]}
        """
        if simplified and vertical:
            computed = self._meridians_simpl_vertical
        elif simplified and not vertical:
            computed = self._meridians_simpl_nonvertical
        elif not simplified and vertical:
            computed = self._meridians_nonsimpl_vertical
        else:
            computed = self._meridians_nonsimpl_nonvertical
        if computed:
            return dict(computed)
        self.fundamental_group(simplified=simplified, vertical=vertical)
        if simplified and vertical:
            return dict(self._meridians_simpl_vertical)
        if simplified and not vertical:
            return dict(self._meridians_group_simpl_nonvertical)
        if not simplified and vertical:
            return dict(self._meridians_nonsimpl_vertical)
        return dict(self._meridians_nonsimpl_nonvertical)

    def braid_monodromy(self, vertical=True):
        r"""
        Return the braid monodromy of the complement of the union
        of affine plane curves in `\CC^2`.

        If there are vertical asymptotes a change of variable is done.

        INPUT:

        - ``vertical`` -- boolean (default: ``True``); if it is ``True``, there
          are no vertical asymptotes, and there are vertical lines, then a
          simplified :func:`braid_monodromy` is computed.

        OUTPUT:

        A braid monodromy with dictionaries identifying strands with components
        and braids with vertical lines.

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x, y + x - 1, x)
            sage: A.braid_monodromy(vertical=False)
            [s1*s0*(s1*s2*s1)^2*s2*(s1^-1*s2^-1)^2*s1^-1*s0^-1*s1^-1,
             s1*s0*(s1*s2)^2*s2*s1^-1*s2^-1*s1^-1*s0^-1*s1^-1,
             s1*s0*s1*s2*(s1*s2^-1)^2*s0*s1*s2*s1*s0*s2^-1*s1^-3*s2*s1^-1*s2^-1*s1^-1*s0^-1*s1^-1,
             s1*s0*s1*s2*s1*s2^-1*s1^4*s2*s1^-1*s2^-1*s1^-1*s0^-1*s1^-1,
             s1*s0*s1*s2*s1*s2^-1*s1^-1*s2*s0^-1*s1^-1]
            sage: A.braid_monodromy(vertical=True)
            [s1*s0*s1*s0^-1*s1^-1*s0, s0^-1*s1*s0*s1^-1*s0, s0^-1*s1^2*s0]
        """
        if vertical:
            computed = self._braid_monodromy_vertical
        else:
            computed = self._braid_monodromy_non_vertical
        if computed is not None:
            return computed
        K = self.base_ring()
        if not K.is_subring(QQbar):
            raise TypeError('the base field is not in QQbar')
        L = self.defining_polynomials()
        bm, dic, dv, d1 = braid_monodromy(prod(L), arrangement=L,
                                          vertical=vertical)
        if vertical:
            self._braid_monodromy_vertical = bm
            self._strands_vertical = dic
            self._vertical_lines_in_braid_mon = dv
        else:
            self._braid_monodromy_non_vertical = bm
            self._strands_nonvertical = dic
        return bm

    def strands(self):
        r"""
        Return the strands for each member of the arrangement.

        OUTPUT:

        A dictionary which associates to the index of each strand
        its associated component if the braid monodromy has been
        calculated with ``vertical=False``.

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x, y + x - 1, x)
            sage: bm = A.braid_monodromy()
            sage: A.strands()
            {0: 2, 1: 1, 2: 0, 3: 0}
        """
        if not self._strands_nonvertical:
            self._braid_monodromy = self.braid_monodromy(vertical=False)
        return self._strands_nonvertical

    def vertical_strands(self):
        r"""
        Return the strands if the braid monodromy has been computed with
        the vertical option.

        OUTPUT:

        A dictionary which associates to the index of each strand
        its associated component if the braid monodromy has been
        calculated with ``vertical=True``.

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x, y + x - 1, x)
            sage: A.vertical_strands()
            {0: 1, 1: 0, 2: 0}
            sage: A.braid_monodromy(vertical=True)
            [s1*s0*s1*s0^-1*s1^-1*s0, s0^-1*s1*s0*s1^-1*s0, s0^-1*s1^2*s0]
        """
        if not self._strands_vertical:
            self.braid_monodromy(vertical=True)
        return self._strands_vertical

    def vertical_lines_in_braid_monodromy(self):
        r"""
        Return the vertical lines in the arrangement.

        OUTPUT:

        A dictionary which associates the index of a braid
        to the index of the vertical line associated to the braid.

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x, y + x - 1, x)
            sage: A.vertical_lines_in_braid_monodromy()
            {1: 2}
            sage: A.braid_monodromy(vertical=True)
            [s1*s0*s1*s0^-1*s1^-1*s0, s0^-1*s1*s0*s1^-1*s0, s0^-1*s1^2*s0]
        """
        if not self._vertical_lines_in_braid_mon:
            self.braid_monodromy(vertical=True)
        return self._vertical_lines_in_braid_mon


class ProjectivePlaneCurveArrangementElement(PlaneCurveArrangementElement):
    """
    An ordered projective plane curve arrangement.
    """
    def __init__(self, parent, curves, check=True):
        """
        Construct an ordered projective plane curve arrangement.

        INPUT:

        - ``parent`` -- the parent :class:`ProjectivePlaneCurveArrangements`

        - ``curves`` -- tuple of curves

        EXAMPLES::

            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: elt = H(x, y, z); elt
            Arrangement (x, y, z) in Projective Space of dimension 2 over Rational Field
            sage: TestSuite(elt).run()
        """
        Element.__init__(self, parent)
        self._curves = tuple(curves)
        if check:
            if not all(isinstance(h, ProjectivePlaneCurve) for h in curves):
                raise ValueError("not all elements are curves")
            if not all(h.ambient_space() is self.parent().ambient_space()
                       for h in curves):
                raise ValueError("not all curves are in the same ambient space")
        self._fundamental_group_nonsimpl = None
        self._fundamental_group_simpl = None
        self._meridians_nonsimpl = None
        self._meridians_simpl = None

    def fundamental_group(self, simplified=True):
        r"""
        Return the fundamental group of the complement of the union
        of an arrangement of projective plane curves
        in the projective plane.

        INPUT:

        - ``simplified`` -- boolean (default: ``True``); set if the group
          is simplified

        OUTPUT: a finitely presented group

        .. NOTE::

           This functionality requires the ``sirocco`` package to be installed.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: H(z).fundamental_group()
            Finitely presented group <  |  >
            sage: H(x*y).fundamental_group()
            Finitely presented group < x |  >
            sage: A = H(y^2 + x*z, y + x - z, x)
            sage: A.fundamental_group().sorted_presentation()
            Finitely presented group < x0, x1 | x1^-1*x0^-1*x1*x0 >
            sage: A.meridians()
            {0: [x1], 1: [x0], 2: [x0^-1*x1^-2]}
            sage: G = A.fundamental_group(simplified=False)
            sage: G.sorted_presentation()
            Finitely presented group
            < x0, x1, x2, x3 | x3^-1*x2^-1*x1^-1*x0^-1, x3^-1*x2^-1*x3*x0*x1*x0^-1,
                               x3^-1*x1^-1*x3*x0*x1*x0^-1*x2^-1*x0^-1*(x2*x0)^2*x1^-1*x0^-1,
                               x3^-1*x0^-1*x3*x0*x1*x0^-1*x2^-1*x0*x2*x0*x1^-1*x0^-1,
                               x2^-1*x0^-1*x2*x0, x1^-1*x0^-1*x1*x0 >
            sage: A.meridians(simplified=False)
            {0: [x1, x2], 1: [x0], 2: [x3]}
            sage: A = H(y^2 + x*z, z, x)
            sage: A.fundamental_group()
            Finitely presented group < x0, x1 | (x1*x0)^2*(x1^-1*x0^-1)^2 >
            sage: A = H(y^2 + x*z, z*x, y)
            sage: A.fundamental_group()
            Finitely presented group
            < x0, x1, x2 | x2*x0*x1*x0^-1*x2^-1*x1^-1,
                           x1*(x2*x0)^2*x2^-1*x1^-1*x0^-1*x2^-1*x0^-1 >
        """
        if simplified:
            computed = self._fundamental_group_simpl
        else:
            computed = self._fundamental_group_nonsimpl
        if computed:
            return computed
        H = self.parent()
        K = self.base_ring()
        R = self.coordinate_ring()
        x, y, z = R.gens()
        if not K.is_subring(QQbar):
            raise TypeError('the base field is not in QQbar')
        C = self.reduce()
        n = len(C)
        infinity = Curve(z)
        infinity_in_C = infinity in C
        if infinity_in_C and n == 1:
            G = FreeGroup(0) / []
            if simplified:
                self._fundamental_group_simpl = G
                self._meridians_simpl = {0: [G.one()]}
            else:
                self._fundamental_group_nonsimpl = G
                self._meridians_nonsimpl = {0: [G.one()]}
            return G
        if infinity_in_C:
            j = C.curves().index(infinity)
            C = H(C.curves()[:j] + C.curves()[j + 1:])
        infinity_divides = False
        for j, c in enumerate(C):
            g = c.defining_polynomial()
            infinity_divides = z.divides(g)
            if infinity_divides:
                h = R(g / z)
                C = H(C.curves()[:j] + (h, ) + C.curves()[j + 1:])
                break
        affine = AffinePlaneCurveArrangements(K, names=('u', 'v'))
        u, v = affine.gens()
        affines = [f.defining_polynomial().subs({x: u, y: v, z: 1}) for f in C]
        changes = any(g.degree(v) < g.degree() > 1 for g in affines)
        while changes:
            affines = [f.subs({u: u + v}) for f in affines]
            changes = any(g.degree(v) < g.degree() > 1 for g in affines)
        C_affine = affine(affines)
        proj = not (infinity_divides or infinity_in_C)
        G = C_affine.fundamental_group(simplified=simplified, vertical=True,
                                       projective=proj)
        dic = C_affine.meridians(simplified=simplified, vertical=True)
        if infinity_in_C:
            dic1 = {}
            for k in range(j):
                dic1[k] = dic[k]
            dic1[j] = dic[n - 1]
            for k in range(j + 1, n):
                dic1[k] = dic[k - 1]
        elif infinity_divides:
            dic1 = {k: dic[k] for k in range(n)}
            dic1[j] += dic[n]
        else:
            dic1 = dic
        if simplified:
            self._fundamental_group_simpl = G
            self._meridians_simpl = dic1
        else:
            self._fundamental_group_nonsimpl = G
            self._meridians_nonsimpl = dic1
        return G

    def meridians(self, simplified=True) -> dict:
        r"""
        Return the meridians of each irreducible component.

        OUTPUT:

        A dictionary which associates the index of each curve with
        its meridians, including the line at infinity if it can be
        computed

        .. NOTE::

           This function requires the ``sirocco`` package to be installed and
           :func:`ProjectivePlaneCurveArrangements.fundamental_group`
           with the same options, where some examples are shown.

        EXAMPLES::

            sage: # needs sirocco
            sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: A = H(y^2 + x*z, y + x - z, x)
            sage: A.fundamental_group().sorted_presentation()
            Finitely presented group < x0, x1 | x1^-1*x0^-1*x1*x0 >
            sage: A.meridians()
            {0: [x1], 1: [x0], 2: [x0^-1*x1^-2]}
            sage: A = H(y^2 + x*z, z, x)
            sage: A.fundamental_group()
            Finitely presented group < x0, x1 | (x1*x0)^2*(x1^-1*x0^-1)^2 >
            sage: A.meridians()
            {0: [x0, x1*x0*x1^-1], 1: [x0^-1*x1^-1*x0^-1], 2: [x1]}
            sage: A = H(y^2 + x*z, z*x, y)
            sage: A.meridians()
            {0: [x0, x2*x0*x2^-1], 1: [x2, x0^-1*x2^-1*x1^-1*x0^-1], 2: [x1]}
            sage: A.fundamental_group()
            Finitely presented group < x0, x1, x2 | x2*x0*x1*x0^-1*x2^-1*x1^-1,
                                                    x1*(x2*x0)^2*x2^-1*x1^-1*x0^-1*x2^-1*x0^-1 >
        """
        if simplified:
            computed = self._meridians_simpl
        else:
            computed = self._meridians_nonsimpl
        if computed:
            return dict(computed)
        self.fundamental_group(simplified=simplified)
        if simplified:
            return dict(self._meridians_simpl)
        return dict(self._meridians_nonsimpl)


class PlaneCurveArrangements(UniqueRepresentation, Parent):
    """
    Plane curve arrangements.

    INPUT:

    - ``base_ring`` -- ring; the base ring

    - ``names`` -- tuple of strings; the variable names

    EXAMPLES::

        sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
        sage: H(x, y^2, x-1, y-1)
        Arrangement (x, y^2, x - 1, y - 1) in Affine Space
        of dimension 2 over Rational Field
    """
    Element = PlaneCurveArrangementElement

    @staticmethod
    def __classcall__(cls, base, names: tuple[str, ...] = ()):
        """
        Normalize the inputs to ensure a unique representation.

        TESTS::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: K = AffinePlaneCurveArrangements(QQ, names=('x', 'y'))
            sage: H is K
            True
        """
        names = normalize_names(len(names), names)
        return super().__classcall__(cls, base, names)

    def __init__(self, base_ring, names: tuple[str, ...] = ()):
        """
        Initialize ``self``.

        TESTS::

            sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: TestSuite(H).run()
        """
        if base_ring not in _Fields:
            raise ValueError('base ring must be a field')
        super().__init__(base_ring, names=names, category=Sets())
        self._embedded = None
        if len(names) == 2:
            self._embedded = 'affine'
        elif len(names) == 3:
            self._embedded = 'projective'

    def coordinate_ring(self):
        """
        Return the coordinate ring.

        OUTPUT: the coordinate ring of the curve arrangement

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.coordinate_ring()
            Multivariate Polynomial Ring in x, y over Rational Field
        """
        return PolynomialRing(self.base_ring(), self.variable_names())

    def change_ring(self, base_ring):
        """
        Return curve arrangements over a different base ring.

        INPUT:

        - ``base_ring`` -- a ring; the new base ring

        OUTPUT:

        A new :class:`PlaneCurveArrangements` instance over the new
        base ring

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.change_ring(RR).base_ring()
            Real Field with 53 bits of precision

        TESTS::

            sage: L.change_ring(QQ) is L
            True
        """
        # return self.__class__(base_ring, names=self.variable_names())
        # line below is ugly but line above does not work
        return self.__reduce__()[1][0](base_ring, names=self.variable_names())

    @abstract_method
    def ambient_space(self):
        """
        Return the ambient space.

        EXAMPLES::

            sage: L.<x, y> = PlaneCurveArrangements(QQ)
            Traceback (most recent call last):
            ...
            NotImplementedError: <abstract method ambient_space at  0x...>
            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.ambient_space()
            Affine Space of dimension 2 over Rational Field
            sage: L.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: L.ambient_space()
            Projective Space of dimension 2 over Rational Field
        """

    def _repr_(self):
        """
        Return a string representation.

        OUTPUT: string

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ);  L
            Curve arrangements in Affine Space of dimension 2 over Rational Field
        """
        return 'Curve arrangements in {}'.format(self.ambient_space())

    def _element_constructor_(self, *args, **kwds):
        """
        Construct an element of ``self``.

        INPUT:

        - ``*args`` -- positional arguments, each defining a curve

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: A = L(x, y); A
            Arrangement (x, y) in Affine Space of dimension 2 over Rational Field
            sage: L([x, y]) == A
            True
            sage: L(Curve(x), Curve(y)) == A
            True
            sage: L(y, x) == A
            False
       """
        if len(args) == 1:
            if not isinstance(args[0], (tuple, list)):
                arg = (args[0], )
            else:
                arg = tuple(args[0])
        else:
            arg = tuple(args)
        ambient_space = self.ambient_space()
        R = ambient_space.coordinate_ring()
        curves: tuple[Any, ...] = ()
        for h in arg:
            try:
                ambient = h.ambient_space()
                if ambient == ambient_space:
                    curves += (h, )
                else:
                    raise TypeError('the curves do not have the same ambient space')
            except AttributeError:
                try:
                    h = R(h)
                    curves += (Curve(h), )
                except TypeError:
                    raise TypeError('elements are not curves')
        return self.element_class(self, curves)

    def _an_element_(self):
        """
        Return an element of ``self``.

        TESTS::

            sage: H.<t, s> = AffinePlaneCurveArrangements(QQ)
            sage: H._an_element_()
            Arrangement (t) in Affine Space of dimension 2 over Rational Field
            sage: H.<t, s, r> = ProjectivePlaneCurveArrangements(QQ)
            sage: H._an_element_()
            Arrangement (t) in Projective Space of dimension 2 over Rational Field
        """
        return self(self.gen(0))

    @cached_method
    def ngens(self):
        """
        Return the number of variables, i.e. 2 or 3, kept for completeness.

        OUTPUT: integer, 2 or 3, depending if the arrangement is projective or affine

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.ngens()
            2
            sage: L.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: L.ngens()
            3
        """
        return len(self.variable_names())

    def gens(self) -> tuple:
        """
        Return the coordinates.

        OUTPUT: a tuple of linear expressions, one for each linear variable

        EXAMPLES::

            sage: L = AffinePlaneCurveArrangements(QQ, ('x', 'y'))
            sage: L.gens()
            (x, y)
            sage: L = ProjectivePlaneCurveArrangements(QQ, ('x', 'y', 'z'))
            sage: L.gens()
            (x, y, z)
        """
        return self.ambient_space().gens()

    def gen(self, i):
        """
        Return the `i`-th coordinate.

        INPUT:

        - ``i`` -- integer

        OUTPUT: a variable

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.gen(1)
            y
            sage: L.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: L.gen(2)
            z
        """
        return self.gens()[i]


class AffinePlaneCurveArrangements(PlaneCurveArrangements):
    """
    Affine curve arrangements.

    INPUT:

    - ``base_ring`` -- ring; the base ring

    - ``names`` -- tuple of strings; the variable names

    EXAMPLES::

        sage: H.<x, y> = AffinePlaneCurveArrangements(QQ)
        sage: H(x, y^2, x-1, y-1)
        Arrangement (x, y^2, x - 1, y - 1) in Affine Space
        of dimension 2 over Rational Field
    """
    Element = AffinePlaneCurveArrangementElement

    def ambient_space(self):
        """
        Return the ambient space.

        EXAMPLES::

            sage: L.<x, y> = AffinePlaneCurveArrangements(QQ)
            sage: L.ambient_space()
            Affine Space of dimension 2 over Rational Field
        """
        return AffineSpace(self.base_ring(), 2, self._names)


class ProjectivePlaneCurveArrangements(PlaneCurveArrangements):
    """
    Projective curve arrangements.

    INPUT:

    - ``base_ring`` -- ring; the base ring

    - ``names`` -- tuple of strings; the variable names

    EXAMPLES::

        sage: H.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
        sage: H(x, y^2, x-z, y-z)
        Arrangement (x, y^2, x - z, y - z) in Projective Space
        of dimension 2 over Rational Field
    """
    Element = ProjectivePlaneCurveArrangementElement

    def ambient_space(self):
        """
        Return the ambient space.

        EXAMPLES::

            sage: L.<x, y, z> = ProjectivePlaneCurveArrangements(QQ)
            sage: L.ambient_space()
            Projective Space of dimension 2 over Rational Field
        """
        return ProjectiveSpace(self.base_ring(), 2, self._names)



# ##########################################################################
#       Copyright (C) 2010 William Stein <wstein@gmail.com>
#  Distributed under the terms of the GNU General Public License (GPL) v2+.
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
# ##########################################################################

cdef class IntList:
    cdef int* _values
    cdef Py_ssize_t _length

    cpdef int prod(self) noexcept
    cpdef int sum(self) noexcept



from typing import Any

class PynacConstant:
    pointer: Any
    _object: Any
    _name: str

    def __cinit__(self, name: str, texname: str, domain_string: str) -> None: ...
    def __dealloc__(self) -> None: ...
    def serial(self) -> int: ...
    def name(self) -> str: ...
    def __repr__(self) -> str: ...
    def expression(self) -> Any: ...



# distutils: libraries = planarity
"""
Wrapper for Boyer's (C) planarity algorithm
"""

cdef extern from "planarity/graphLib.h":
    ctypedef struct vertexRec:
        int link[2]
        int index
    ctypedef vertexRec * vertexRecP

    ctypedef struct edgeRec:
        int link[2]
        int neighbor
    ctypedef edgeRec * edgeRecP

    ctypedef struct BM_graph:
        vertexRecP V
        edgeRecP E
        int N
    ctypedef BM_graph * graphP

    cdef int OK, EMBEDFLAGS_PLANAR, NONEMBEDDABLE, NOTOK

    cdef graphP gp_New()
    cdef void gp_Free(graphP *pGraph)
    cdef int gp_EnsureVertexCapacity(graphP theGraph, int N)
    cdef int gp_AddEdge(graphP theGraph, int u, int ulink, int v, int vlink)
    cdef int gp_Embed(graphP theGraph, int embedFlags)
    cdef int gp_SortVertices(graphP theGraph)


def is_planar(g, kuratowski=False, set_pos=False, set_embedding=False, immutable=None):
    r"""
    Check whether ``g`` is planar using Boyer's planarity algorithm.

    If ``kuratowski`` is ``False``, returns ``True`` if ``g`` is planar,
    ``False`` otherwise.  If ``kuratowski`` is ``True``, returns a tuple, first
    entry is a boolean (whether or not the graph is planar) and second entry is
    a Kuratowski subgraph, i.e. an edge subdivision of `K_5` or `K_{3,3}` (if
    not planar) or ``None`` (if planar).  Also, will set an ``_embedding``
    attribute for the graph ``g`` if ``set_embedding`` is set to ``True``.

    INPUT:

    - ``kuratowski`` -- boolean (default: ``False``); when set to ``True``,
      return a tuple of a boolean and either ``None`` or a Kuratowski subgraph
      (i.e. an edge subdivision of `K_5` or `K_{3,3}`). When set to ``False``,
      returns ``True`` if ``g`` is planar, ``False`` otherwise.

    - ``set_pos`` -- boolean (default: ``False``); whether to use Schnyder's
      algorithm to determine and set positions

    - ``set_embedding`` -- boolean (default: ``False``); whether to record the
      combinatorial embedding returned (see
      :meth:`~sage.graphs.generic_graph.GenericGraph.get_embedding`)

    - ``immutable`` -- boolean (default: ``None``); whether to create a
      mutable/immutable graph. ``immutable=None`` (default) means that
      the graph and the Kuratowski subgraph will behave the same way.
      This parameter is ignored when ``kuratowski=False``.

    EXAMPLES::

        sage: G = graphs.DodecahedralGraph()
        sage: from sage.graphs.planarity import is_planar
        sage: is_planar(G)
        True
        sage: Graph('@').is_planar()
        True

    TESTS:

    We try checking the planarity of all graphs on 7 or fewer
    vertices.  In fact, to try to track down a segfault, we do it
    twice. ::

        sage: # long time, needs networkx
        sage: import networkx.generators.atlas
        sage: atlas_graphs = [Graph(i) for i in networkx.generators.atlas.graph_atlas_g()]
        sage: a = [i for i in [1..1252] if atlas_graphs[i].is_planar()]
        sage: b = [i for i in [1..1252] if atlas_graphs[i].is_planar()]
        sage: a == b
        True

    There were some problems with ``set_pos`` stability in the past,
    so let's check if this runs without exception::

        sage: for i, g in enumerate(atlas_graphs):      # long time                     # needs networkx
        ....:     if (not g.is_connected() or i == 0):
        ....:         continue
        ....:     _ = g.is_planar(set_embedding=True, set_pos=True)

        Argument saving::

            sage: G = Graph([(1, 2)])
            sage: for set_embedding, set_pos in ((True,True), (True,False), (False, True), (False, False)):
            ....:     G = Graph([(1, 2)])
            ....:     assert is_planar(G, set_embedding=set_embedding, set_pos=set_pos)
            ....:     assert (hasattr(G, '_embedding') and G._embedding is not None) == set_embedding, (set_embedding, set_pos)
            ....:     assert (hasattr(G, '_pos') and G._pos is not None) == set_pos, (set_embedding, set_pos)

    Check the behavior of parameter ``immutable``::

        sage: G = graphs.PetersenGraph()
        sage: G.is_planar(kuratowski=True)
        (False, Kuratowski subgraph of (Petersen graph): Graph on 9 vertices)
        sage: G.is_planar(kuratowski=True)[1].is_immutable()
        False
        sage: G.is_planar(kuratowski=True, immutable=True)[1].is_immutable()
        True
        sage: G = G.copy(immutable=True)
        sage: G.is_planar(kuratowski=True)[1].is_immutable()
        True
        sage: G.is_planar(kuratowski=True, immutable=False)[1].is_immutable()
        False
    """
    g._scream_if_not_simple()
    if set_pos and not g.is_connected():
        raise ValueError("is_planar() cannot set vertex positions for a disconnected graph")

    # First take care of a trivial cases
    if not g.size():
        # There are no edges
        if set_embedding:
            g._embedding = {v: [] for v in g}
        return (True, None) if kuratowski else True
    if g.order() == 2 and g.is_connected():
        # P_2 is too small to be triangulated
        u, v = list(g)
        if set_embedding:
            g._embedding = {u: [v], v: [u]}
        if set_pos:
            g._pos = {u: [0, 0], v: [0, 1]}
        return (True, None) if kuratowski else True

    # Create to and from mappings to relabel vertices to the set {1,...,n}
    # (planarity 3 uses 1-based array indexing, with 0 representing NIL)
    cdef int i
    cdef list listto = list(g)
    cdef dict ffrom = {vvv: i for i, vvv in enumerate(listto, start=1)}
    cdef dict to = {i: vvv for i, vvv in enumerate(listto, start=1)}

    cdef graphP theGraph
    theGraph = gp_New()
    cdef int status
    status = gp_EnsureVertexCapacity(theGraph, g.order())
    if status != OK:
        raise RuntimeError("gp_EnsureVertexCapacity status is not ok")
    for u, v in g.edge_iterator(labels=False):
        status = gp_AddEdge(theGraph, ffrom[u], 0, ffrom[v], 0)
        if status == NOTOK:
            raise RuntimeError("gp_AddEdge status is not ok")
        elif status == NONEMBEDDABLE:
            # We now know that the graph is nonplanar.
            if not kuratowski:
                gp_Free(&theGraph)
                return False
            # With just the current edges, we have a nonplanar graph,
            # so to isolate a Kuratowski subgraph, just keep going.
            break

    status = gp_Embed(theGraph, EMBEDFLAGS_PLANAR)

    if status == NOTOK:
        raise RuntimeError("status is not ok")

    gp_SortVertices(theGraph)

    if status == NONEMBEDDABLE:
        # Kuratowski subgraph isolator
        if not kuratowski:
            gp_Free(&theGraph)
            return False
        g_dict = {}
        for i in range(1, theGraph.N + 1):
            linked_list = []
            j = theGraph.V[i].link[1]
            while j:
                linked_list.append(to[theGraph.E[j].neighbor])
                j = theGraph.E[j].link[1]
            if linked_list:
                g_dict[to[i]] = linked_list
        if immutable is None:
            immutable = g.is_immutable()
        gp_Free(&theGraph)
        G = g.__class__(data=g_dict, weighted=g._weighted,
                        loops=g.allows_loops(),
                        multiedges=g.allows_multiple_edges(),
                        name="Kuratowski subgraph of (%s)" % g.name(),
                        immutable=immutable)
        if g.get_pos():
            G.set_pos({u: g._pos[u] for u in g_dict})
        return (False, G)

    if set_pos or set_embedding:
        emb_dict = {}
        for i in range(1, theGraph.N + 1):
            linked_list = []
            j = theGraph.V[i].link[1]
            while j:
                linked_list.append(to[theGraph.E[j].neighbor])
                j = theGraph.E[j].link[1]
            emb_dict[to[i]] = linked_list
        if set_embedding:
            g._embedding = emb_dict
        if set_pos:
            g.layout(layout='planar', save_pos=True, on_embedding=emb_dict)

    gp_Free(&theGraph)
    if kuratowski:
        return (True, None)
    return True



# sage.doctest: needs sage.libs.flint
"""
Degeneracy maps
"""

# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from . import morphism


class DegeneracyMap(morphism.HeckeModuleMorphism_matrix):
    """
    A degeneracy map between Hecke modules of different levels.

    EXAMPLES:

    We construct a number of degeneracy maps::

        sage: M = ModularSymbols(33)
        sage: d = M.degeneracy_map(11)
        sage: d
        Hecke module morphism degeneracy map corresponding to f(q) |--> f(q)
        defined by the matrix
        [ 1  0  0]
        [ 0  0  1]
        [ 0  0 -1]
        [ 0  1 -1]
        [ 0  0  1]
        [ 0 -1  1]
        [-1  0  0]
        [-1  0  0]
        [-1  0  0]
        Domain: Modular Symbols space of dimension 9 for Gamma_0(33) of weight ...
        Codomain: Modular Symbols space of dimension 3 for Gamma_0(11) of weight ...
        sage: d.t()
        1
        sage: d = M.degeneracy_map(11,3)
        sage: d.t()
        3

    The parameter d must be a divisor of the quotient of the two levels::

        sage: d = M.degeneracy_map(11,2)
        Traceback (most recent call last):
        ...
        ValueError: the level of self (=33) must be a divisor or multiple
        of level (=11) and t (=2) must be a divisor of the quotient

    Degeneracy maps can also go from lower level to higher level::

        sage: M.degeneracy_map(66,2)
        Hecke module morphism degeneracy map corresponding to f(q) |--> f(q^2)
        defined by the matrix
        [ 2  0  0  0  0  0  1  0  0  0  1 -1  0  0  0 -1  1  0  0  0  0  0  0  0 -1]
        [ 0  0  1 -1  0 -1  1  0 -1  2  0  0  0 -1  0  0 -1  1  2 -2  0  0  0 -1  1]
        [ 0  0  1  0  0  0  0  0  1  0  0  0  1  0  0  0 -1  1  0  0 -1  1  0  0  0]
        [ 0  0  0  0  0  0  0  0  0  2 -1  0  0  1  0  0 -1  1  0  0  1  0 -1 -1  1]
        [ 0 -1  0  0  1  0  0  0  0  0  0  1  0  0  1  1 -1  0  0 -1  0  0  0  0  0]
        [ 0  0  0  0  0  0  0  1 -1  0  0  2 -1  0  0  1  0  0  0 -1  0 -1  1 -1  1]
        [ 0  0  0  0  1 -1  0  1 -1  0  0  0  0  0 -1  2  0  0  0  0  1  0  1  0  0]
        [ 0  0  1  0  0  0  0  0  0  0  0  0  0  0  0  0  0  1  0  0  0  1  1  0  0]
        [ 0  0  0  0  0  0  0  0  0  0 -1  0  0  0  0  0  0  0  0  1  1  1  0  0  0]
        Domain: Modular Symbols space of dimension 9 for Gamma_0(33) of weight ...
        Codomain: Modular Symbols space of dimension 25 for Gamma_0(66) of weight ...
    """
    def __init__(self, matrix, domain, codomain, t):
        r"""
        Initialise a degeneracy map.

        EXAMPLES::

            sage: D = ModularSymbols(Gamma0(100)).degeneracy_map(2,5); D
            Hecke module morphism degeneracy map corresponding to f(q) |--> f(q^5) defined by the matrix
            31 x 1 dense matrix over Rational Field
            Domain: Modular Symbols space of dimension 31 for Gamma_0(100) of weight ...
            Codomain: Modular Symbols space of dimension 1 for Gamma_0(2) of weight ...
            sage: D == loads(dumps(D))
            True
        """
        self.__t = t
        H = domain.Hom(codomain)
        if t == 1:
            pow = ""
        else:
            pow = "^%s" % t
        name = "degeneracy map corresponding to f(q) |--> f(q%s)" % (pow)
        morphism.HeckeModuleMorphism_matrix.__init__(self, H, matrix, name)

    def t(self):
        """
        Return the divisor of the quotient of the two levels
        associated to the degeneracy map.

        EXAMPLES::

            sage: M = ModularSymbols(33)
            sage: d = M.degeneracy_map(11,3)
            sage: d.t()
            3
            sage: d = M.degeneracy_map(11,1)
            sage: d.t()
            1
        """
        return self.__t



# sage.doctest: needs sage.libs.pari
r"""
Canonical forms and automorphisms for linear codes over finite fields

We implemented the algorithm described in [Feu2009]_ which computes, a unique
code (canonical form) in the equivalence class of a given
linear code `C \leq \GF{q}^n`. Furthermore, this algorithm will return the
automorphism group of `C`, too. You will find more details about the algorithm
in the documentation of the class
:class:`~sage.coding.codecan.autgroup_can_label.LinearCodeAutGroupCanLabel`.

The equivalence of codes is modeled as a group action by the group
`G = {\GF{q}^*}^n \rtimes (Aut(\GF{q}) \times S_n)` on the set
of subspaces of `\GF{q}^n` . The group `G` will be called the
semimonomial group of degree `n`.

The algorithm is started by initializing the class
:class:`~sage.coding.codecan.autgroup_can_label.LinearCodeAutGroupCanLabel`.
When the object gets available, all computations are already finished and
you can access the relevant data using the member functions:

* :meth:`~sage.coding.codecan.autgroup_can_label.LinearCodeAutGroupCanLabel.get_canonical_form`

* :meth:`~sage.coding.codecan.autgroup_can_label.LinearCodeAutGroupCanLabel.get_transporter`

* :meth:`~sage.coding.codecan.autgroup_can_label.LinearCodeAutGroupCanLabel.get_autom_gens`

People do also use some weaker notions of equivalence, namely
**permutational** equivalence and monomial equivalence (**linear** isometries).
These can be seen as the subgroups `S_n` and `{\GF{q}^*}^n \rtimes S_n` of `G`.
If you are interested in one of these notions, you can just pass
the optional parameter ``algorithm_type``.

A second optional parameter ``P`` allows you to restrict the
group of permutations `S_n` to a subgroup which respects the coloring given
by ``P``.

AUTHORS:

- Thomas Feulner (2012-11-15): initial version

EXAMPLES::

    sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
    sage: C = codes.HammingCode(GF(3), 3).dual_code()
    sage: P = LinearCodeAutGroupCanLabel(C)
    sage: P.get_canonical_form().generator_matrix()
    [1 0 0 0 0 1 1 1 1 1 1 1 1]
    [0 1 0 1 1 0 0 1 1 2 2 1 2]
    [0 0 1 1 2 1 2 1 2 1 2 0 0]
    sage: LinearCode(P.get_transporter()*C.generator_matrix()) == P.get_canonical_form()
    True
    sage: A = P.get_autom_gens()
    sage: all(LinearCode(a*C.generator_matrix()) == C for a in A)
    True
    sage: P.get_autom_order() == GL(3, GF(3)).order()
    True

If the dimension of the dual code is smaller, we will work on this code::

    sage: C2 = codes.HammingCode(GF(3), 3)
    sage: P2 = LinearCodeAutGroupCanLabel(C2)
    sage: P2.get_canonical_form().parity_check_matrix() == P.get_canonical_form().generator_matrix()
    True

There is a specialization of this algorithm to pass a coloring on the
coordinates. This is just a list of lists, telling the algorithm which
columns do share the same coloring::

    sage: C = codes.HammingCode(GF(4, 'a'), 3).dual_code()
    sage: P = LinearCodeAutGroupCanLabel(C, P=[ [0], [1], list(range(2, C.length())) ])
    sage: P.get_autom_order()
    864
    sage: A = [a.get_perm() for a in P.get_autom_gens()]
    sage: H = SymmetricGroup(21).subgroup(A)
    sage: H.orbits()
    ((1,),
     (2,),
     (3, 5, 4),
     (6, 19, 16, 21, 9, 10, 15, 8, 20, 11, 14, 13, 7, 12, 18, 17))

We can also restrict the group action to linear isometries::

    sage: P = LinearCodeAutGroupCanLabel(C, algorithm_type='linear')
    sage: P.get_autom_order() == GL(3, GF(4, 'a')).order()
    True

and to the action of the symmetric group only::

    sage: P = LinearCodeAutGroupCanLabel(C, algorithm_type='permutational')
    sage: P.get_autom_order() == C.permutation_automorphism_group().order()
    True
"""
# ****************************************************************************
#       Copyright (C) 2012 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.coding.codecan.codecan import PartitionRefinementLinearCode
from sage.combinat.permutation import Permutation
from sage.arith.misc import factorial


def _cyclic_shift(n, p):
    r"""
    If ``p`` is a list of pairwise distinct coordinates in ``range(n)``,
    then this function returns the cyclic shift of
    the coordinates contained in ``p``, when acting on a vector of length `n`.

    Note that the domain of a ``Permutation`` is ``range(1, n+1)``.

    EXAMPLES::

        sage: from sage.coding.codecan.autgroup_can_label import _cyclic_shift
        sage: p = _cyclic_shift(10, [2,7,4,1]); p
        [1, 3, 8, 4, 2, 6, 7, 5, 9, 10]

    We prove that the action is as expected::

        sage: t = list(range(10))
        sage: p.action(t)
        [0, 2, 7, 3, 1, 5, 6, 4, 8, 9]
    """
    x = list(range(1, n + 1))
    for i in range(1, len(p)):
        x[p[i - 1]] = p[i] + 1
    x[p[len(p) - 1]] = p[0] + 1
    return Permutation(x)


class LinearCodeAutGroupCanLabel:
    r"""
    Canonical representatives and automorphism group computation for linear
    codes over finite fields.

    There are several notions of equivalence for linear codes:
    Let `C`, `D` be linear codes of length `n` and dimension `k`.
    The codes `C` and `D` are said to be

    - permutational equivalent, if there is some permutation `\pi \in S_n`
      such that `(c_{\pi(0)}, \ldots, c_{\pi(n-1)}) \in D` for all `c \in C`.

    - linear equivalent, if there is some permutation `\pi \in S_n` and a
      vector `\phi \in {\GF{q}^*}^n` of units of length `n` such that
      `(c_{\pi(0)} \phi_0^{-1}, \ldots, c_{\pi(n-1)} \phi_{n-1}^{-1}) \in D`
      for all `c \in C`.

    - semilinear equivalent, if there is some permutation `\pi \in S_n`, a
      vector `\phi` of units of length `n` and a field automorphism `\alpha`
      such that
      `(\alpha(c_{\pi(0)}) \phi_0^{-1}, \ldots, \alpha( c_{\pi(n-1)}) \phi_{n-1}^{-1} ) \in D`
      for all `c \in C`.

    These are group actions. This class provides an algorithm that will compute
    a unique representative `D` in the orbit of the given linear code `C`.
    Furthermore, the group element `g` with `g * C = D` and the automorphism
    group of `C` will be computed as well.

    There is also the possibility to restrict the permutational part of this
    action to a Young subgroup of `S_n`. This could be achieved by passing a
    partition `P` (as a list of lists) of the set `\{0, \ldots, n-1\}`. This is
    an option which is also available in the computation of a canonical form of
    a graph, see :meth:`sage.graphs.generic_graph.GenericGraph.canonical_label`.

    EXAMPLES::

        sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
        sage: C = codes.HammingCode(GF(3), 3).dual_code()
        sage: P = LinearCodeAutGroupCanLabel(C)
        sage: P.get_canonical_form().generator_matrix()
        [1 0 0 0 0 1 1 1 1 1 1 1 1]
        [0 1 0 1 1 0 0 1 1 2 2 1 2]
        [0 0 1 1 2 1 2 1 2 1 2 0 0]
        sage: LinearCode(P.get_transporter()*C.generator_matrix()) == P.get_canonical_form()
        True
        sage: a = P.get_autom_gens()[0]
        sage: (a*C.generator_matrix()).echelon_form() == C.generator_matrix().echelon_form()
        True
        sage: P.get_autom_order() == GL(3, GF(3)).order()
        True
    """

    def __init__(self, C, P=None, algorithm_type='semilinear'):
        """
        See :class:`LinearCodeAutGroupCanLabel`.

        INPUT:

        - ``C`` -- a linear code

        - ``P`` -- (optional) a coloring of the coordinates i.e. a partition
          (list of disjoint lists) of [0 , ..., C.length()-1 ]

        - ``algorithm_type`` -- (optional) defines the acting group, either

            * ``permutational``

            * ``linear``

            * ``semilinear``

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(2), 3).dual_code()
            sage: P = LinearCodeAutGroupCanLabel(C)
            sage: P.get_canonical_form().generator_matrix()
            [1 0 0 0 1 1 1]
            [0 1 0 1 0 1 1]
            [0 0 1 1 1 1 0]
            sage: P2 = LinearCodeAutGroupCanLabel(C, P=[[0,3,5],[1,2,4,6]],
            ....:      algorithm_type='permutational')
            sage: P2.get_canonical_form().generator_matrix()
            [1 0 0 1 0 1 1]
            [0 1 0 1 1 0 1]
            [0 0 1 0 1 1 1]
        """
        from sage.groups.semimonomial_transformations.semimonomial_transformation_group import SemimonomialTransformationGroup
        from sage.coding.linear_code import LinearCode, AbstractLinearCode

        if not isinstance(C, AbstractLinearCode):
            raise TypeError("%s is not a linear code" % C)

        self.C = C
        mat = C.generator_matrix()
        F = mat.base_ring()
        S = SemimonomialTransformationGroup(F, mat.ncols())

        if P is None:
            P = [list(range(mat.ncols()))]

        pos2P = [-1] * mat.ncols()
        for i in range(len(P)):
            P[i].sort(reverse=True)
            for x in P[i]:
                pos2P[x] = i

        col_list = mat.columns()
        nz = [i for i in range(mat.ncols()) if not col_list[i].is_zero()]
        z = [(pos2P[i], i) for i in range(mat.ncols()) if col_list[i].is_zero()]
        z.sort()
        z = [i for (p, i) in z]

        normalization_factors = [F.one()] * mat.ncols()
        if algorithm_type == "permutational":
            for c in col_list:
                c.set_immutable()
        else:
            for x in nz:
                normalization_factors[x] = col_list[x][(col_list[x].support())[0]]
                col_list[x] = normalization_factors[x] ** (-1) * col_list[x]
                col_list[x].set_immutable()

        normalization = S(v=normalization_factors)
        normalization_inverse = normalization ** (-1)
        col_set = list({col_list[y] for y in nz})
        col2pos = []
        col2P = []
        for c in col_set:
            X = [(pos2P[y], y) for y in range(mat.ncols()) if col_list[y] == c]
            X.sort()
            col2pos.append([b for _, b in X])
            col2P.append([a for a, _ in X])

        zipped = sorted(zip(col2P, col_set, col2pos))

        col2P = [qty for qty, c, pos in zipped]
        col_set = [c for qty, c, pos in zipped]
        col2pos = [pos for qty, c, pos in zipped]
        P_refined = []
        p = [0]
        act_qty = col2P[0]
        for i in range(1, len(col_set)):
            if act_qty == col2P[i]:
                p.append(i)
            else:
                P_refined.append(p)
                p = [i]
                act_qty = col2P[i]
        P_refined.append(p)
        # now we can start the main algorithm:
        from sage.matrix.constructor import matrix

        if len(col_set) >= 2 * mat.nrows():
            # the dimension of the code is smaller or equal than
            # the dimension of the dual code
            # in this case we work with the code itself.
            pr = PartitionRefinementLinearCode(len(col_set),
                                               matrix(col_set).transpose(),
                                               P=P_refined,
                                               algorithm_type=algorithm_type)

            # this command allows you some advanced debugging
            # it prints the backtrack tree -> must be activated when installing
            # pr._latex_view(title='MyTitle') #this will provide you some visual representation of what is going on

            can_transp = pr.get_transporter()
            can_col_set = pr.get_canonical_form().columns()
            self._PGammaL_autom_gens = self._compute_PGammaL_automs(
                pr.get_autom_gens(),
                normalization,
                normalization_inverse, col2pos
            )
            self._PGammaL_autom_size = pr.get_autom_order_permutation()
            self._PGammaL_autom_size *= pr.get_autom_order_inner_stabilizer()
            self._full_autom_order = self._PGammaL_autom_size
            self._PGammaL_autom_size /= (len(self.C.base_ring()) - 1)
        elif mat.nrows() == len(col_set):
            # it could happen (because of the recursive call, see `else`) that
            # the canonical form is the identity matrix
            n = len(col_set)
            from sage.modules.free_module import VectorSpace
            can_col_set = VectorSpace(F, n).gens()
            A = []
            self._full_autom_order = 1
            S_short = SemimonomialTransformationGroup(F, n)
            can_transp = S_short.one()
            if algorithm_type != "permutational":
                # linear or semilinear
                if algorithm_type == "semilinear":
                    A.append(S_short(autom=F.hom([F.gen() ** F.characteristic()])))
                    self._full_autom_order *= F.degree()
                for p in P_refined:
                    m = [F.one()] * n
                    m[p[0]] = F.gen()
                    A.append(S_short(v=m))
                self._full_autom_order *= (len(F) - 1) ** n
            for p in P_refined:
                if len(p) > 1:
                    A.append(S_short(perm=_cyclic_shift(n, p[:2])))
                    if len(p) > 2:
                        # cyclic shift of all elements
                        A.append(S_short(perm=_cyclic_shift(n, p)))
                    self._full_autom_order *= factorial(len(p))
            self._PGammaL_autom_size = self._full_autom_order / (len(F) - 1)
            self._PGammaL_autom_gens = self._compute_PGammaL_automs(
                A, normalization, normalization_inverse, col2pos)
        else:
            # use the dual code for the computations
            # this might have zero columns or multiple columns, hence
            # we call this algorithm again.
            short_dual_code = LinearCode(matrix(col_set).transpose()).dual_code()
            agcl = LinearCodeAutGroupCanLabel(short_dual_code,
                                              P=P_refined,
                                              algorithm_type=algorithm_type)
            can_transp = agcl.get_transporter()
            can_transp.invert_v()
            can_col_set = agcl.get_canonical_form().parity_check_matrix().columns()
            A = agcl.get_autom_gens()
            for a in A:
                a.invert_v()
            self._PGammaL_autom_gens = self._compute_PGammaL_automs(
                A, normalization, normalization_inverse, col2pos)
            self._PGammaL_autom_size = agcl.get_PGammaL_order()
            self._full_autom_order = agcl.get_autom_order()

        canonical_form = matrix(F, mat.ncols(), mat.nrows())

        perm = [-1] * mat.ncols()
        mult = [F.one()] * mat.ncols()

        for i in range(len(can_col_set)):
            img = can_transp.get_perm()(i + 1)
            for j in col2pos[img - 1]:
                pos = P[pos2P[j]].pop()
                canonical_form[pos] = can_col_set[i]
                mult[pos] = can_transp.get_v()[i]
                perm[pos] = j + 1

        it = iter(z)
        for p in P:
            while p:
                pos = p.pop()
                perm[pos] = next(it) + 1

        self._canonical_form = LinearCode(canonical_form.transpose())
        self._transporter = S(perm=Permutation(perm), v=mult, autom=can_transp.get_autom()) * normalization
        self._trivial_autom_gens, a = self._compute_trivial_automs(
            normalization,
            normalization_inverse, z, [pos2P[x] for x in z],
            zero_column_case=True)
        self._full_autom_order *= a

        for i in range(len(col2P)):
            if len(col2P[i]) > 1:
                A, a = self._compute_trivial_automs(normalization,
                                                    normalization_inverse,
                                                    col2pos[i], col2P[i])
                self._full_autom_order *= a
                self._trivial_autom_gens += A

    @staticmethod
    def _compute_trivial_automs(normalization, normalization_inverse, col2pos, col2P, zero_column_case=False):
        """
        In order to call the algorithm described in
        :class:`PartitionRefinementLinearCode` we remove
        zero columns and multiple occurrences of the same column (up to
        normalization).

        This function computes a set of generators for the allowed permutations
        of such a block of equal columns (depending on the partition).
        If ``zero_column_case==True`` we also add a generator for the
        multiplication by units. Furthermore, we return the order of this group.

        INPUT:

        - ``normalization`` -- an element in the semimonomial transformation group `S`

        - ``normalization_inverse`` -- the inverse of ``normalization``

        - ``col2pos`` -- list of disjoint indices in ``range(n)``

        - ``col2P`` -- an increasing list of integers, with
          ``len(col2P) == len(col2pos)`` with ``col2P[i] == col2P[j]`` if and
          only if the indices ``col2pos[i]`` and ``col2pos[j]`` are in the
          same partition

        - ``zero_column_case`` -- boolean; set to ``True`` iff we are dealing
          with the zero column

        OUTPUT:

        - list of generators in `S`

        - the order of this group

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: S = SemimonomialTransformationGroup(GF(3), 10)
            sage: s = S.one()
            sage: col2pos = [1,4,2,3,5]
            sage: col2P = [1,1,3,3,3]
            sage: LinearCodeAutGroupCanLabel._compute_trivial_automs(s, s, col2pos, col2P, True)
            ([((1, 2, 1, 1, 1, 1, 1, 1, 1, 1); (), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1), ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1); (2,5), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1), ((1, 1, 2, 1, 1, 1, 1, 1, 1, 1); (), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1), ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1); (3,4), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1), ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1); (3,4,6), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1)], 384)
        """
        S = normalization.parent()
        F = S.base_ring()
        n = S.degree()

        beg = 0
        if zero_column_case:
            aut_order = (len(F) - 1) ** len(col2P)
        else:
            aut_order = 1
        A = []
        while beg < len(col2P):
            if zero_column_case:
                mult = [F.one()] * n
                mult[col2pos[beg]] = F.primitive_element()
                A.append(S(v=mult))
            P_indx = col2P[beg]
            j = beg + 1
            while j < len(col2P) and col2P[j] == P_indx:
                j += 1

            if j - beg > 1:
                aut_order *= factorial(j - beg)
                # we append a transposition of the first two elements
                A.append(normalization_inverse *
                         S(perm=_cyclic_shift(n, col2pos[beg:beg + 2])) *
                         normalization)
                if j - beg > 2:
                    # we append a cycle on all elements
                    A.append(normalization_inverse *
                             S(perm=_cyclic_shift(n, col2pos[beg:j])) *
                             normalization)
            beg = j
        return A, aut_order

    @staticmethod
    def _compute_PGammaL_automs(gens, normalization, normalization_inverse, col2pos):
        """
        In order to call the algorithm described in
        :class:`sage.coding.codecan.codecan.PartitionRefinementLinearCode` we removed
        zero columns and multiple occurrences of the same column (up to normalization).
        This function lifts the generators ``gens`` which were returned to their full length.

        INPUT:

        - ``gens`` -- list of semimonomial transformation group elements of length `m`

        - ``normalization`` -- a semimonomial transformation of length `n`

        - ``normalization_inverse`` -- the inverse of ``normalization``

        - ``col2pos`` -- a partition of ``range(n)`` into `m` disjoint sets,
          given as a list of lists. The elements `g` in ``gens`` are only
          allowed to permute entries of ``col2pos`` of equal length.

        OUTPUT:

        - a list of semimonomial transformations containing
          ``normalization`` which are the lifts of the elements in ``gens``

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: S = SemimonomialTransformationGroup(GF(3), 10)
            sage: T = SemimonomialTransformationGroup(GF(3), 5)
            sage: s = S.one()
            sage: col2pos = [[1,4,2,3,5], [0],[6],[8],[7,9]]
            sage: gens = [T(perm=Permutation([1,3,4,2,5]), v=[2,1,1,1,1])]
            sage: LinearCodeAutGroupCanLabel._compute_PGammaL_automs(gens, s, s, col2pos)
            [((1, 2, 2, 2, 2, 2, 1, 1, 1, 1); (1,7,9), Ring endomorphism of Finite Field of size 3
              Defn: 1 |--> 1)]
        """
        S = normalization.parent()
        n = S.degree()
        A = []
        for g in gens:
            perm = list(range(1, n + 1))
            mult = [S.base_ring().one()] * n
            short_perm = g.get_perm()
            short_mult = g.get_v()
            for i in range(len(col2pos)):
                c = col2pos[i]
                img_iter = iter(col2pos[short_perm(i + 1) - 1])
                for x in c:
                    perm[x] = next(img_iter) + 1
                    mult[x] = short_mult[i]
            A.append(normalization_inverse * S(perm=perm, v=mult, autom=g.get_autom()) * normalization)
        return A

    def get_canonical_form(self):
        """
        Return the canonical orbit representative we computed.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(3), 3).dual_code()
            sage: CF1 = LinearCodeAutGroupCanLabel(C).get_canonical_form()
            sage: s = SemimonomialTransformationGroup(GF(3), C.length()).an_element()
            sage: C2 = LinearCode(s*C.generator_matrix())
            sage: CF2 = LinearCodeAutGroupCanLabel(C2).get_canonical_form()
            sage: CF1 == CF2
            True
        """
        return self._canonical_form

    def get_transporter(self):
        """
        Return the element which maps the code to its canonical form.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(2), 3).dual_code()
            sage: P = LinearCodeAutGroupCanLabel(C)
            sage: g = P.get_transporter()
            sage: D = P.get_canonical_form()
            sage: (g*C.generator_matrix()).echelon_form() == D.generator_matrix().echelon_form()
            True
        """
        return self._transporter

    def get_autom_gens(self):
        """
        Return a generating set for the automorphism group of the code.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(2), 3).dual_code()
            sage: A = LinearCodeAutGroupCanLabel(C).get_autom_gens()
            sage: Gamma = C.generator_matrix().echelon_form()
            sage: all((g*Gamma).echelon_form() == Gamma for g in A)
            True
        """
        return self._PGammaL_autom_gens + self._trivial_autom_gens

    def get_autom_order(self):
        """
        Return the size of the automorphism group of the code.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(2), 3).dual_code()
            sage: LinearCodeAutGroupCanLabel(C).get_autom_order()
            168
        """
        return self._full_autom_order

    def get_PGammaL_gens(self):
        r"""
        Return the set of generators translated to the group `P\Gamma L(k,q)`.

        There is a geometric point of view of code equivalence. A linear
        code is identified with the multiset of points in the finite projective
        geometry `PG(k-1, q)`. The equivalence of codes translates to the
        natural action of `P\Gamma L(k,q)`. Therefore, we may interpret the
        group as a subgroup of `P\Gamma L(k,q)` as well.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(4, 'a'), 3).dual_code()
            sage: A = LinearCodeAutGroupCanLabel(C).get_PGammaL_gens()
            sage: Gamma = C.generator_matrix()
            sage: N = [ x.monic() for x in Gamma.columns() ]
            sage: all((g[0]*n.apply_map(g[1])).monic() in N for n in N for g in A)
            True
        """
        Gamma = self.C.generator_matrix()
        res = []
        for a in self._PGammaL_autom_gens:
            B = Gamma.solve_left(a * Gamma, check=True)
            res.append([B, a.get_autom()])

        return res

    def get_PGammaL_order(self):
        r"""
        Return the size of the automorphism group as a subgroup of
        `P\Gamma L(k,q)`.

        There is a geometric point of view of code equivalence. A linear
        code is identified with the multiset of points in the finite projective
        geometry `PG(k-1, q)`. The equivalence of codes translates to the
        natural action of `P\Gamma L(k,q)`. Therefore, we may interpret the
        group as a subgroup of `P\Gamma L(k,q)` as well.

        EXAMPLES::

            sage: from sage.coding.codecan.autgroup_can_label import LinearCodeAutGroupCanLabel
            sage: C = codes.HammingCode(GF(4, 'a'), 3).dual_code()
            sage: LinearCodeAutGroupCanLabel(C).get_PGammaL_order() == GL(3, GF(4, 'a')).order()*2/3
            True
        """
        return self._PGammaL_autom_size



# distutils: libraries = flint
# distutils: depends = flint/mag.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void mag_init(mag_t x) noexcept
    void mag_clear(mag_t x) noexcept
    void mag_swap(mag_t x, mag_t y) noexcept
    mag_ptr _mag_vec_init(slong n) noexcept
    void _mag_vec_clear(mag_ptr v, slong n) noexcept
    slong mag_allocated_bytes(const mag_t x) noexcept
    void mag_zero(mag_t res) noexcept
    void mag_one(mag_t res) noexcept
    void mag_inf(mag_t res) noexcept
    bint mag_is_special(const mag_t x) noexcept
    bint mag_is_zero(const mag_t x) noexcept
    bint mag_is_inf(const mag_t x) noexcept
    bint mag_is_finite(const mag_t x) noexcept
    void mag_init_set(mag_t res, const mag_t x) noexcept
    void mag_set(mag_t res, const mag_t x) noexcept
    void mag_set_d(mag_t res, double x) noexcept
    void mag_set_ui(mag_t res, ulong x) noexcept
    void mag_set_fmpz(mag_t res, const fmpz_t x) noexcept
    void mag_set_d_lower(mag_t res, double x) noexcept
    void mag_set_ui_lower(mag_t res, ulong x) noexcept
    void mag_set_fmpz_lower(mag_t res, const fmpz_t x) noexcept
    void mag_set_d_2exp_fmpz(mag_t res, double x, const fmpz_t y) noexcept
    void mag_set_fmpz_2exp_fmpz(mag_t res, const fmpz_t x, const fmpz_t y) noexcept
    void mag_set_ui_2exp_si(mag_t res, ulong x, slong y) noexcept
    void mag_set_d_2exp_fmpz_lower(mag_t res, double x, const fmpz_t y) noexcept
    void mag_set_fmpz_2exp_fmpz_lower(mag_t res, const fmpz_t x, const fmpz_t y) noexcept
    double mag_get_d(const mag_t x) noexcept
    double mag_get_d_log2_approx(const mag_t x) noexcept
    void mag_get_fmpq(fmpq_t res, const mag_t x) noexcept
    void mag_get_fmpz(fmpz_t res, const mag_t x) noexcept
    void mag_get_fmpz_lower(fmpz_t res, const mag_t x) noexcept
    bint mag_equal(const mag_t x, const mag_t y) noexcept
    int mag_cmp(const mag_t x, const mag_t y) noexcept
    int mag_cmp_2exp_si(const mag_t x, slong y) noexcept
    void mag_min(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_max(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_print(const mag_t x) noexcept
    void mag_fprint(FILE * file, const mag_t x) noexcept
    char * mag_dump_str(const mag_t x) noexcept
    int mag_load_str(mag_t x, const char * str) noexcept
    int mag_dump_file(FILE * stream, const mag_t x) noexcept
    int mag_load_file(mag_t x, FILE * stream) noexcept
    void mag_randtest(mag_t res, flint_rand_t state, slong expbits) noexcept
    void mag_randtest_special(mag_t res, flint_rand_t state, slong expbits) noexcept
    void mag_add(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_add_ui(mag_t res, const mag_t x, ulong y) noexcept
    void mag_add_lower(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_add_ui_lower(mag_t res, const mag_t x, ulong y) noexcept
    void mag_add_2exp_fmpz(mag_t res, const mag_t x, const fmpz_t e) noexcept
    void mag_add_ui_2exp_si(mag_t res, const mag_t x, ulong y, slong e) noexcept
    void mag_sub(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_sub_lower(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_mul_2exp_si(mag_t res, const mag_t x, slong y) noexcept
    void mag_mul_2exp_fmpz(mag_t res, const mag_t x, const fmpz_t y) noexcept
    void mag_mul(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_mul_ui(mag_t res, const mag_t x, ulong y) noexcept
    void mag_mul_fmpz(mag_t res, const mag_t x, const fmpz_t y) noexcept
    void mag_mul_lower(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_mul_ui_lower(mag_t res, const mag_t x, ulong y) noexcept
    void mag_mul_fmpz_lower(mag_t res, const mag_t x, const fmpz_t y) noexcept
    void mag_addmul(mag_t z, const mag_t x, const mag_t y) noexcept
    void mag_div(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_div_ui(mag_t res, const mag_t x, ulong y) noexcept
    void mag_div_fmpz(mag_t res, const mag_t x, const fmpz_t y) noexcept
    void mag_div_lower(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_inv(mag_t res, const mag_t x) noexcept
    void mag_inv_lower(mag_t res, const mag_t x) noexcept
    void mag_fast_init_set(mag_t x, const mag_t y) noexcept
    void mag_fast_zero(mag_t res) noexcept
    bint mag_fast_is_zero(const mag_t x) noexcept
    void mag_fast_mul(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_fast_addmul(mag_t z, const mag_t x, const mag_t y) noexcept
    void mag_fast_add_2exp_si(mag_t res, const mag_t x, slong e) noexcept
    void mag_fast_mul_2exp_si(mag_t res, const mag_t x, slong e) noexcept
    void mag_pow_ui(mag_t res, const mag_t x, ulong e) noexcept
    void mag_pow_fmpz(mag_t res, const mag_t x, const fmpz_t e) noexcept
    void mag_pow_ui_lower(mag_t res, const mag_t x, ulong e) noexcept
    void mag_pow_fmpz_lower(mag_t res, const mag_t x, const fmpz_t e) noexcept
    void mag_sqrt(mag_t res, const mag_t x) noexcept
    void mag_sqrt_lower(mag_t res, const mag_t x) noexcept
    void mag_rsqrt(mag_t res, const mag_t x) noexcept
    void mag_rsqrt_lower(mag_t res, const mag_t x) noexcept
    void mag_hypot(mag_t res, const mag_t x, const mag_t y) noexcept
    void mag_root(mag_t res, const mag_t x, ulong n) noexcept
    void mag_log(mag_t res, const mag_t x) noexcept
    void mag_log_lower(mag_t res, const mag_t x) noexcept
    void mag_neg_log(mag_t res, const mag_t x) noexcept
    void mag_neg_log_lower(mag_t res, const mag_t x) noexcept
    void mag_log_ui(mag_t res, ulong n) noexcept
    void mag_log1p(mag_t res, const mag_t x) noexcept
    void mag_exp(mag_t res, const mag_t x) noexcept
    void mag_exp_lower(mag_t res, const mag_t x) noexcept
    void mag_expinv(mag_t res, const mag_t x) noexcept
    void mag_expinv_lower(mag_t res, const mag_t x) noexcept
    void mag_expm1(mag_t res, const mag_t x) noexcept
    void mag_exp_tail(mag_t res, const mag_t x, ulong N) noexcept
    void mag_binpow_uiui(mag_t res, ulong m, ulong n) noexcept
    void mag_geom_series(mag_t res, const mag_t x, ulong N) noexcept
    void mag_const_pi(mag_t res) noexcept
    void mag_const_pi_lower(mag_t res) noexcept
    void mag_atan(mag_t res, const mag_t x) noexcept
    void mag_atan_lower(mag_t res, const mag_t x) noexcept
    void mag_cosh(mag_t res, const mag_t x) noexcept
    void mag_cosh_lower(mag_t res, const mag_t x) noexcept
    void mag_sinh(mag_t res, const mag_t x) noexcept
    void mag_sinh_lower(mag_t res, const mag_t x) noexcept
    void mag_fac_ui(mag_t res, ulong n) noexcept
    void mag_rfac_ui(mag_t res, ulong n) noexcept
    void mag_bin_uiui(mag_t res, ulong n, ulong k) noexcept
    void mag_bernoulli_div_fac_ui(mag_t res, ulong n) noexcept
    void mag_polylog_tail(mag_t res, const mag_t z, slong s, ulong d, ulong N) noexcept
    void mag_hurwitz_zeta_uiui(mag_t res, ulong s, ulong a) noexcept

from .mag_macros cimport *
